# MIT License
#
# Copyright (c) 2018 Haoxintong
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Residual Attention network, implemented in Gluon."""

from mxnet.gluon import nn
from mxnet.gluon.model_zoo.vision.resnet import BottleneckV2
from ..nn.basic_blocks import FrBase


__all__ = ["AttentionNet", "AttentionNetFace",
           "get_attention_net", "get_attention_face",
           "attention_net56", "attention_net92", "attention_net128",
           "attention_net164", "attention_net236", "attention_net452"
           ]


class AttentionBlock(nn.HybridBlock):
    def __init__(self, channels, out_size, stage, p=1, t=2, r=1, **kwargs):
        r"""Residual Attention Block from
        `"Residual Attention Network for Image Classification"
        <https://arxiv.org/abs/1704.06904>`_ paper.

        Parameters
        ----------
        channels: int.
            Number of output channels.
        out_size: int.
            Size of the output feature map, now it only supports square shape.
        stage: int.
            Stage described in Figure 2.
        p: int.
            Number of pre-processing Residual Units before split into trunk branch and mask branch.
        t: int.
            Number of Residual Units in trunk branch.
        r: int.
            Number of Residual Units between adjacent pooling layer in the mask branch.
        kwargs:
        """
        super().__init__(**kwargs)
        with self.name_scope():
            self.pre = nn.HybridSequential()
            for i in range(p):
                self.pre.add(BottleneckV2(channels, 1, prefix='pre_%d_' % i))

            self.trunk_branch = nn.HybridSequential()
            for i in range(t):
                self.trunk_branch.add(BottleneckV2(channels, 1, prefix='trunk_%d_' % i))

            self.mask_branch = _MaskBlock(channels, r, out_size, stage, prefix='mask_')

            self.post = nn.HybridSequential()
            for i in range(p):
                self.post.add(BottleneckV2(channels, 1, prefix='post_%d_' % i))

    def hybrid_forward(self, F, x, *args, **kwargs):
        x = self.pre(x)
        mask = self.mask_branch(x)
        trunk = self.trunk_branch(x)
        out = (1 + mask) * trunk
        out = self.post(out)
        return out


class _UpSampleBlock(nn.HybridBlock):
    def __init__(self, out_size, **kwargs):
        super().__init__(**kwargs)
        self._size = out_size

    def hybrid_forward(self, F, x, *args, **kwargs):
        return F.contrib.BilinearResize2D(x, height=self._size, width=self._size)


class _MaskBlock(nn.HybridBlock):
    def __init__(self, channels, r, out_size, stage, **kwargs):
        super().__init__(**kwargs)
        self._stage = stage
        with self.name_scope():
            self._make_layers(channels, r, stage, out_size)

    def _make_layers(self, channels, r, stage, out_size):
        if stage <= 1:
            self.down_sample_1 = nn.MaxPool2D(3, 2, 1)
            self.down_res_unit_1 = nn.HybridSequential()
            for i in range(r):
                self.down_res_unit_1.add(BottleneckV2(channels, 1, prefix="down_res1_%d_" % i))
            self.skip_connection_1 = BottleneckV2(channels, 1)

            self.up_res_unit_1 = nn.HybridSequential()
            for i in range(r):
                self.up_res_unit_1.add(BottleneckV2(channels, 1, prefix="up_res1_%d_" % i))
            self.up_sample_1 = _UpSampleBlock(out_size)
            out_size = out_size // 2

        if stage <= 2:
            self.down_sample_2 = nn.MaxPool2D(3, 2, 1)
            self.down_res_unit_2 = nn.HybridSequential()
            for i in range(r):
                self.down_res_unit_2.add(BottleneckV2(channels, 1, prefix="down_res2_%d_" % i))
            self.skip_connection_2 = BottleneckV2(channels, 1)

            self.up_res_unit_2 = nn.HybridSequential()
            for i in range(r):
                self.up_res_unit_2.add(BottleneckV2(channels, 1, prefix="up_res2_%d_" % i))
            self.up_sample_2 = _UpSampleBlock(out_size)
            out_size = out_size // 2

        if stage <= 3:
            self.down_sample_3 = nn.MaxPool2D(3, 2, 1)
            self.down_res_unit_3 = nn.HybridSequential()
            for i in range(r):
                self.down_res_unit_3.add(BottleneckV2(channels, 1, prefix="down_res3_%d_" % i))

            self.up_res_unit_3 = nn.HybridSequential()
            for i in range(r):
                self.up_res_unit_3.add(BottleneckV2(channels, 1, prefix="up_res3_%d_" % i))
            self.up_sample_3 = _UpSampleBlock(out_size)

        self.output = nn.HybridSequential()
        self.output.add(nn.BatchNorm(),
                        nn.Activation('relu'),
                        nn.Conv2D(channels, kernel_size=1, strides=1, use_bias=False),
                        nn.BatchNorm(),
                        nn.Activation('relu'),
                        nn.Conv2D(channels, kernel_size=1, strides=1, use_bias=False),
                        nn.Activation('sigmoid')
                        )

    def hybrid_forward(self, F, x, *args, **kwargs):
        if self._stage <= 1:
            x_down1 = self.down_sample_1(x)
            x_down1 = self.down_res_unit_1(x_down1)
            residual_1 = self.skip_connection_1(x_down1)
        else:
            x_down1 = x
            residual_1 = 0

        if self._stage <= 2:
            x_down2 = self.down_sample_2(x_down1)
            x_down2 = self.down_res_unit_2(x_down2)
            residual_2 = self.skip_connection_2(x_down2)
        else:
            x_down2 = x
            residual_2 = 0

        if self._stage <= 3:
            x_down3 = self.down_sample_3(x_down2)
            x_down3 = self.down_res_unit_3(x_down3)

            x_up3 = self.up_res_unit_3(x_down3)
            x_up3 = self.up_sample_3(x_up3)
        else:
            raise ValueError("param stage should be a number not larger than 3!")

        if self._stage <= 2:
            x_up2 = x_up3 + residual_2
            x_up2 = self.up_res_unit_2(x_up2)
            x_up2 = self.up_sample_2(x_up2)
        else:
            x_up2 = x_up3

        if self._stage <= 1:
            x_up1 = x_up2 + residual_1
            x_up1 = self.up_res_unit_1(x_up1)
            x_up1 = self.up_sample_1(x_up1)
        else:
            x_up1 = x_up2
        out = self.output(x_up1)
        return out


class AttentionNet(nn.HybridBlock):
    r"""AttentionNet Model from
    `"Residual Attention Network for Image Classification"
    <https://arxiv.org/abs/1704.06904>`_ paper.

    Parameters
    ----------
    classes: int.
        Number of classification classes.
    modules: list.
        The number of Attention Module in each stage.
    p: int.
        Number of pre-processing Residual Units before split into trunk branch and mask branch.
    t: int.
        Number of Residual Units in trunk branch.
    r: int.
        Number of Residual Units between adjacent pooling layer in the mask branch.
    kwargs:

    """

    def __init__(self, classes, modules, p, t, r, **kwargs):
        super().__init__(**kwargs)
        assert len(modules) == 3
        with self.name_scope():
            self.features = nn.HybridSequential()
            # 112x112
            self.features.add(nn.Conv2D(64, 3, 2, 1, use_bias=False))
            self.features.add(nn.BatchNorm())
            self.features.add(nn.Activation('relu'))

            # 56x56
            self.features.add(nn.MaxPool2D(3, 2, 1))
            self.features.add(BottleneckV2(256, 1, True, 64))
            for _ in range(modules[0]):
                self.features.add(AttentionBlock(256, 56, 1, p, t, r))

            # 28x28
            self.features.add(BottleneckV2(512, 2, True, 256))
            for _ in range(modules[1]):
                self.features.add(AttentionBlock(512, 28, 2, p, t, r))

            # 14x14
            self.features.add(BottleneckV2(1024, 2, True, 512))
            for _ in range(modules[2]):
                self.features.add(AttentionBlock(1024, 14, 3, p, t, r))

            # 7x7
            self.features.add(BottleneckV2(2048, 2, True, 1024),
                              BottleneckV2(2048, 1),
                              BottleneckV2(2048, 1))

            # 2048
            self.features.add(nn.BatchNorm(),
                              nn.Activation('relu'),
                              nn.GlobalAvgPool2D(),
                              nn.Flatten())

            # classes
            self.output = nn.Dense(classes)

    def hybrid_forward(self, F, x, *args, **kwargs):
        x = self.features(x)
        x = self.output(x)
        return x


class AttentionNetFace(FrBase):
    r"""AttentionNet Model for input 112x112.

    Parameters
    ----------
    classes: int.
        Number of classification classes.
    modules: list.
        The number of Attention Module in each stage.
    p: int.
        Number of pre-processing Residual Units before split into trunk branch and mask branch.
    t: int.
        Number of Residual Units in trunk branch.
    r: int.
        Number of Residual Units between adjacent pooling layer in the mask branch.
    embedding_size : int
        Units of embedding layer.
    weight_norm : bool, default False
        Whether use weight norm in NormDense layer.
    feature_norm : bool, default False
        Whether use features norm in NormDense layer.
    need_cls_layer : bool, default True
        Whether use NormDense layer.Normally it depends on your loss function.
        When you use Softmax, ArcLoss or based on Softmax loss, you need to set it to True.
        When you only need embedding output, like you are predicting or training with triplet loss,
        you need to set it to False.


    """

    def __init__(self, classes, modules, p, t, r,
                 weight_norm=False, feature_norm=False, embedding_size=512,
                 need_cls_layer=True, **kwargs):
        super().__init__(classes, embedding_size, weight_norm, feature_norm, need_cls_layer, **kwargs)
        assert len(modules) == 3
        with self.name_scope():
            self.features = nn.HybridSequential()
            # 112x112
            self.features.add(nn.Conv2D(64, 3, 1, 1, use_bias=False))
            self.features.add(nn.BatchNorm())
            self.features.add(nn.Activation('relu'))

            # 56x56
            self.features.add(BottleneckV2(256, 2, True, 64))
            for _ in range(modules[0]):
                self.features.add(AttentionBlock(256, 56, 1, p, t, r))

            # 28x28
            self.features.add(BottleneckV2(512, 2, True, 256))
            for _ in range(modules[1]):
                self.features.add(AttentionBlock(512, 28, 2, p, t, r))

            # 14x14
            self.features.add(BottleneckV2(1024, 2, True, 512))
            for _ in range(modules[2]):
                self.features.add(AttentionBlock(1024, 14, 3, p, t, r))

            # 8x8
            self.features.add(BottleneckV2(2048, 2, True, 1024),
                              BottleneckV2(2048, 1),
                              BottleneckV2(2048, 1))

            # 2048
            self.features.add(nn.BatchNorm(),
                              nn.Activation('relu'),
                              nn.GlobalAvgPool2D(),
                              nn.Flatten())
            # embedding
            self.features.add(nn.Dense(embedding_size, use_bias=False),
                              nn.BatchNorm(scale=False, center=False),
                              nn.PReLU())


# Specification ([p, t, r], [stage1, stage2, stage3])
# The hyper-parameters are based on paper section 4.1. The number of layers can be calculated by 36m+20
# where m is the number of Attention Module in each stage when `p, t, r = 1, 2, 1`.
attention_net_spec = {56: ([1, 2, 1], [1, 1, 1]),
                      92: ([1, 2, 1], [1, 2, 3]),
                      128: ([1, 2, 1], [3, 3, 3]),
                      164: ([1, 2, 1], [4, 4, 4]),
                      236: ([1, 2, 1], [6, 6, 6]),
                      452: ([2, 4, 3], [6, 6, 6])}


# Constructor
def get_attention_net(classes, num_layers, **kwargs):
    r"""AttentionNet Model from
    `"Residual Attention Network for Image Classification"
    <https://arxiv.org/abs/1704.06904>`_ paper.

    Parameters
    ----------
    classes : int,
        Number of classification classes.
    num_layers : int
        Numbers of layers. Options are 56, 92, 128, 164, 236, 452.
    """
    assert num_layers in attention_net_spec, \
        "Invalid number of layers: %d. Options are %s" % (
            num_layers, str(attention_net_spec.keys()))
    ptr, modules = attention_net_spec[num_layers]
    assert len(ptr) == len(modules) == 3
    p, t, r = ptr
    net = AttentionNet(classes, modules, p, t, r, **kwargs)
    return net


def get_attention_face(classes=-1, num_layers=128, embedding_size=512, need_cls_layer=True, **kwargs):
    r"""AttentionNet Model for 112x112 face images from
    `"Residual Attention Network for Image Classification"
    <https://arxiv.org/abs/1704.06904>`_ paper.

    Parameters
    ----------
    classes : int, -1
        Number of classification classes.
    num_layers : int, 128
        Numbers of layers. Options are 56, 92, 128, 164, 236, 452.
    embedding_size: int, 256
        Feature dimensions of the embedding layers.
    need_cls_layer : bool, default True
        Whether to use NormDense output layer.
    """
    assert num_layers in attention_net_spec, \
        "Invalid number of layers: %d. Options are %s" % (
            num_layers, str(attention_net_spec.keys()))
    ptr, modules = attention_net_spec[num_layers]
    assert len(ptr) == len(modules) == 3
    p, t, r = ptr
    net = AttentionNetFace(classes, modules, p, t, r, embedding_size=embedding_size, need_cls_layer=need_cls_layer, **kwargs)
    return net


def attention_net56(classes=-1, need_cls_layer=True, **kwargs):
    r"""AttentionNet 56 Model for face recognition.

    Parameters
    ----------
    classes : int, -1
       Number of classification classes.
    need_cls_layer : bool, default True
       Whether to use NormDense output layer.
    """
    return get_attention_face(classes, 56, need_cls_layer=need_cls_layer, **kwargs)


def attention_net92(classes=-1, need_cls_layer=True, **kwargs):
    r"""AttentionNet 92 Model for face recognition.

    Parameters
    ----------
    classes : int, -1
       Number of classification classes.
    need_cls_layer : bool, default True
       Whether to use NormDense output layer.

    """
    return get_attention_face(classes, 92, need_cls_layer=need_cls_layer, **kwargs)


def attention_net128(classes=-1, need_cls_layer=True, **kwargs):
    r"""AttentionNet 128 Model for face recognition.

    Parameters
    ----------
    classes : int, -1
       Number of classification classes.
    need_cls_layer : bool, default True
       Whether to use NormDense output layer.
    """
    return get_attention_face(classes, 128, need_cls_layer=need_cls_layer, **kwargs)


def attention_net164(classes=-1, need_cls_layer=True, **kwargs):
    r"""AttentionNet 164 Model for face recognition.

    Parameters
    ----------
    classes : int, -1
       Number of classification classes.
    need_cls_layer : bool, default True
       Whether to use NormDense output layer.
    """
    return get_attention_face(classes, 164, need_cls_layer=need_cls_layer, **kwargs)


def attention_net236(classes=-1, need_cls_layer=True, **kwargs):
    r"""AttentionNet 236 Model for face recognition.

    Parameters
    ----------
    classes : int, -1
       Number of classification classes.
    need_cls_layer : bool, default True
       Whether to use NormDense output layer.
    """
    return get_attention_face(classes, 236, need_cls_layer=need_cls_layer, **kwargs)


def attention_net452(classes=-1, need_cls_layer=True, **kwargs):
    r"""AttentionNet 452 Model for face recognition.

    Parameters
    ----------
    classes : int, -1
       Number of classification classes.
    need_cls_layer : bool, default True
       Whether to use NormDense output layer.
    """
    return get_attention_face(classes, 452, need_cls_layer=need_cls_layer, **kwargs)



# MIT License
#
# Copyright (c) 2018 Haoxintong
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
""""""
from mxnet.gluon import nn
from ..nn.basic_blocks import FrBase, SELayer

__all__ = ["MobileFaceNet",
           "get_mobile_facenet",
           "get_mobile_facenet_re"
           ]


def _make_conv(stage_index, channels=1, kernel=1, stride=1, pad=0,
               num_group=1, active=True):
    out = nn.HybridSequential(prefix='stage%d_' % stage_index)
    out.add(nn.Conv2D(channels, kernel, stride, pad, groups=num_group, use_bias=False))
    out.add(nn.BatchNorm(scale=True))
    if active:
        out.add(nn.PReLU())
    return out


def _make_bottleneck(stage_index, layers, channels, stride, t, in_channels=0, use_se=False):
    layer = nn.HybridSequential(prefix='stage%d_' % stage_index)
    with layer.name_scope():
        layer.add(Bottleneck(in_channels=in_channels, channels=channels, t=t, stride=stride, use_se=use_se))
        for _ in range(layers - 1):
            layer.add(Bottleneck(channels, channels, t, 1, use_se=use_se))
    return layer


class Bottleneck(nn.HybridBlock):

    def __init__(self, in_channels, channels, t, stride, use_se=False, **kwargs):
        super(Bottleneck, self).__init__(**kwargs)
        self.use_shortcut = stride == 1 and in_channels == channels
        with self.name_scope():
            self.out = nn.HybridSequential()
            self.out.add(_make_conv(0, in_channels * t),
                         _make_conv(1, in_channels * t, kernel=3, stride=stride,
                                    pad=1, num_group=in_channels * t),
                         _make_conv(2, channels, active=False))
            if use_se:
                self.out.add(SELayer(channels, channels))

    def hybrid_forward(self, F, x, **kwargs):
        out = self.out(x)
        if self.use_shortcut:
            out = F.elemwise_add(out, x)
        return out


class MobileFaceNet(FrBase):
    """Mobile FaceNet"""
    def __init__(self, classes=-1, use_se=False, weight_norm=False,
                 feature_norm=False, need_cls_layer=True, **kwargs):
        super(MobileFaceNet, self).__init__(classes, 128, weight_norm,
                                            feature_norm, need_cls_layer, **kwargs)
        with self.name_scope():
            self.features = nn.HybridSequential(prefix='feature_')
            with self.features.name_scope():
                self.features.add(_make_conv(0, 64, kernel=3, stride=2, pad=1),
                                  _make_conv(0, 64, kernel=3, stride=1, pad=1, num_group=64))

                self.features.add(
                    _make_bottleneck(1, layers=5, channels=64, stride=2, t=2, in_channels=64, use_se=use_se),
                    _make_bottleneck(2, layers=1, channels=128, stride=2, t=4, in_channels=64, use_se=use_se),
                    _make_bottleneck(3, layers=6, channels=128, stride=1, t=2, in_channels=128, use_se=use_se),
                    _make_bottleneck(4, layers=1, channels=128, stride=2, t=4, in_channels=128, use_se=use_se),
                    _make_bottleneck(5, layers=2, channels=128, stride=1, t=2, in_channels=128, use_se=use_se))

                self.features.add(_make_conv(6, 512),
                                  _make_conv(6, 512, kernel=7, num_group=512, active=False),
                                  nn.Conv2D(128, 1, use_bias=False),
                                  nn.BatchNorm(scale=False, center=False),
                                  nn.Flatten())


class MobileFaceNet_re(FrBase):
    """Mobile FaceNet re"""
    def __init__(self, classes=-1, use_se=False, weight_norm=False,
                 feature_norm=False, need_cls_layer=True, **kwargs):
        super(MobileFaceNet_re, self).__init__(classes, 160, weight_norm,
                                               feature_norm, need_cls_layer, **kwargs)
        with self.name_scope():
            self.features = nn.HybridSequential(prefix='feature_')
            with self.features.name_scope():
                self.features.add(_make_conv(0, 32, kernel=3, stride=2, pad=1),
                                  # _make_conv(0, 64, kernel=3, stride=1, pad=1, num_group=64)
                                  )

                self.features.add(
                    _make_bottleneck(1, layers=1, channels=24, stride=2, t=3, in_channels=32, use_se=use_se),
                    _make_bottleneck(2, layers=2, channels=32, stride=2, t=3, in_channels=24, use_se=use_se),
                    _make_bottleneck(3, layers=3, channels=64, stride=1, t=3, in_channels=32, use_se=use_se),
                    _make_bottleneck(4, layers=2, channels=96, stride=2, t=3, in_channels=64, use_se=use_se),
                    _make_bottleneck(5, layers=2, channels=160, stride=1, t=3, in_channels=96, use_se=use_se))

                self.features.add(_make_conv(6, 320),
                                  _make_conv(6, 320, kernel=7, num_group=320, active=False),
                                  nn.Conv2D(160, 1, use_bias=False),
                                  nn.BatchNorm(scale=False, center=False),
                                  nn.Flatten())


def get_mobile_facenet(classes=-1, need_cls_layer=True, **kwargs):
    r"""
    Parameters
    ----------
    classes : int, -1
       Number of classification classes.
    need_cls_layer : bool, default True
       Whether to use NormDense output layer.
    """
    return MobileFaceNet(classes=classes, need_cls_layer=need_cls_layer, **kwargs)


def get_mobile_facenet_re(classes=-1, need_cls_layer=True, **kwargs):
    r"""
    Parameters
    ----------
    classes : int, -1
       Number of classification classes.
    need_cls_layer : bool, default True
       Whether to use NormDense output layer.
    """
    return MobileFaceNet_re(classes=classes, need_cls_layer=need_cls_layer, **kwargs)



# MIT License
#
# Copyright (c) 2018 Haoxintong
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Face Recognition Dataset"""
import os
import pickle
import mxnet as mx
import numpy as np
from mxnet.gluon.data import Dataset
from mxnet.gluon.data import RecordFileDataset
from mxnet import image, recordio

__all__ = ["FRValDataset",
           "FRTrainRecordDataset",
           ]


class FRTrainRecordDataset(RecordFileDataset):
    r"""A dataset wrapping over a rec serialized file provided by InsightFace Repo.

    Parameters
    ----------
    name : str. Name of val dataset.
    root : str. Path to face folder. Default is '$(HOME)/mxnet/datasets/face'
    transform : function, default None
        A user defined callback that transforms each sample. For example:
        ::

            transform=lambda data, label: (data.astype(np.float32)/255, label)
    """

    def __init__(self, name, root=os.path.expanduser('~/.mxnet/datasets/face'), flag=1, transform=None):
        super().__init__(os.path.join(root, name, "train.rec"))
        prop = open(os.path.join(root, name, "property"), "r").read().strip().split(',')
        self._flag = flag
        self._transform = transform

        assert len(prop) == 3
        self.num_classes = int(prop[0])
        self.image_size = [int(prop[1]), int(prop[2])]

    def __getitem__(self, idx):
        while True:
            record = super().__getitem__(idx)
            header, img = recordio.unpack(record)
            if _check_valid_image(img):
                decoded_img = image.imdecode(img, self._flag)
            else:
                idx = np.random.randint(low=0, high=len(self))
                continue
            if self._transform is not None:
                return self._transform(decoded_img, header.label)
            return decoded_img, header.label


def _check_valid_image(s):
    return False if len(s) == 0 else True


class FRValDataset(Dataset):
    r"""A dataset wrapping over a pickle serialized (.bin) file provided by InsightFace Repo.

    Parameters
    ----------
    name : str.
        Name of val dataset.
    root : str.
        Path to face folder. Default is '$(HOME)/mxnet/datasets/face'
    transform : callable, default None.
        A function that takes data and transforms them:
        ::

            transform = lambda data: data.astype(np.float32)/255

    """

    def __init__(self, name, root=os.path.expanduser('~/.mxnet/datasets/face'), transform=None):
        super().__init__()
        self._transform = transform
        self.name = name
        with open(os.path.join(root, "{}.bin".format(name)), 'rb') as f:
            self.bins, self.issame_list = pickle.load(f, encoding='iso-8859-1')

        self._do_encode = not isinstance(self.bins[0], np.ndarray)

    def __getitem__(self, idx):
        img0 = self._decode(self.bins[2 * idx])
        img1 = self._decode(self.bins[2 * idx + 1])

        issame = 1 if self.issame_list[idx] else 0

        if self._transform is not None:
            img0 = self._transform(img0)
            img1 = self._transform(img1)

        return (img0, img1), issame

    def __len__(self):
        return len(self.issame_list)

    def _decode(self, im):
        if self._do_encode:
            im = im.encode("iso-8859-1")
        return mx.image.imdecode(im)



# MIT License
#
# Copyright (c) 2018 Haoxintong
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
""" Models for face recognition"""
from .mobile_facenet import *
from .attention_net import *
from .se_resnet import *


_models = {
    'l_se_resnet18v2': se_resnet18_v2,
    'l_se_resnet50v2': se_resnet50_v2,
    'l_se_resnet101v2': se_resnet101_v2,
    'mobilefacenet': get_mobile_facenet,
    'attention_facenet': get_attention_face,
    'attentionfacenet56': attention_net56,
    'attentionfacenet92': attention_net92,
}


def get_model(name, **kwargs):
    """Returns a model by name.

    Parameters
    ----------
    name : str
        Name of the model.
    classes : int
        Number of classes for the output layer.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.


    Returns
    -------
    HybridBlock
        The model.
    """
    name = name.lower()
    if name not in _models:
        err_str = '"%s" is not among the following model list:\n\t' % (name)
        err_str += '%s' % ('\n\t'.join(sorted(_models.keys())))
        raise ValueError(err_str)
    net = _models[name](**kwargs)
    return net


def get_model_list():
    """Get the entire list of model names in model_zoo.

    Returns
    -------
    list of str
        Entire list of model names in model_zoo.
    """
    return _models.keys()




# @File  : se_resnet.py
# @Author: X.Yang
# @Contact : pistonyang@gmail.com
# @Date  : 18-11-2
"""SE_ResNetV2 is not from origin paper.
It's proposed from  ArcFace paper as SE-LResNet.
`"ArcFace: Additive Angular Margin Loss for Deep Face Recognition"
<https://arxiv.org/pdf/1801.07698.pdf>`_ paper.
"""
from mxnet.gluon import nn
from mxnet.gluon.nn import HybridBlock
from mxnet.gluon.model_zoo.vision.resnet import _conv3x3
from ..nn.basic_blocks import FrBase, SELayer


__all__ = ['get_se_resnet', 'se_resnet18_v2', 'se_resnet34_v2',
           'se_resnet50_v2', 'se_resnet101_v2', 'se_resnet152_v2']


class SE_BottleneckV2(HybridBlock):
    r"""Bottleneck V2 from
    `"Identity Mappings in Deep Residual Networks"
    <https://arxiv.org/abs/1603.05027>`_ paper.
    This is used for ResNet V2 for 50, 101, 152 layers.

    Parameters
    ----------
    channels : int
        Number of output channels.
    stride : int
        Stride size.
    downsample : bool, default False
        Whether to downsample the input.
    in_channels : int, default 0
        Number of input channels. Default is 0, to infer from the graph.
    """

    def __init__(self, channels, stride, downsample=False, in_channels=0, **kwargs):
        super(SE_BottleneckV2, self).__init__(**kwargs)
        self.bn1 = nn.BatchNorm()
        self.prelu1 = nn.PReLU()
        self.conv1 = nn.Conv2D(channels // 4, kernel_size=1, strides=1, use_bias=False)
        self.bn2 = nn.BatchNorm()
        self.prelu2 = nn.PReLU()
        self.conv2 = _conv3x3(channels // 4, stride, channels // 4)
        self.bn3 = nn.BatchNorm()
        self.prelu3 = nn.PReLU()
        self.conv3 = nn.Conv2D(channels, kernel_size=1, strides=1, use_bias=False)
        if downsample:
            self.downsample = nn.Conv2D(channels, 1, stride, use_bias=False,
                                        in_channels=in_channels)
        else:
            self.downsample = None

        self.se = SELayer(channels, channels)

    def hybrid_forward(self, F, x):
        residual = x
        x = self.bn1(x)
        x = self.prelu1(x)
        if self.downsample:
            residual = self.downsample(x)
        x = self.conv1(x)

        x = self.bn2(x)
        x = self.prelu2(x)
        x = self.conv2(x)

        x = self.bn3(x)
        x = self.prelu3(x)
        x = self.conv3(x)

        x = self.se(x)
        return x + residual


class SE_ResNetV2(FrBase):
    r"""ResNet V2 model from
    `"Identity Mappings in Deep Residual Networks"
    <https://arxiv.org/abs/1603.05027>`_ paper.

    Parameters
    ----------
    block : HybridBlock
        Class for the residual block. Options are BasicBlockV1, BottleneckV1.
    layers : list of int
        Numbers of layers in each block
    channels : list of int
        Numbers of channels in each block. Length should be one larger than layers list.
    classes : int, default 1000
        Number of classification classes.
    thumbnail : bool, default False
        Enable thumbnail.
    """

    def __init__(self, block, layers, channels, classes=-1, thumbnail=False,
                 embedding_size=512, weight_norm=False, feature_norm=False,
                 need_cls_layer=True, **kwargs):
        super(SE_ResNetV2, self).__init__(classes, embedding_size, weight_norm,
                                          feature_norm, need_cls_layer, **kwargs)
        assert len(layers) == len(channels) - 1
        with self.name_scope():
            self.features = nn.HybridSequential(prefix='')
            self.features.add(nn.BatchNorm(scale=False, center=False))
            if thumbnail:
                self.features.add(_conv3x3(channels[0], 1, 0))
            else:
                self.features.add(nn.Conv2D(channels[0], 3, 1, 1, use_bias=False))
                self.features.add(nn.BatchNorm())
                self.features.add(nn.PReLU())
                self.features.add(nn.MaxPool2D(3, 2, 1))

            in_channels = channels[0]
            for i, num_layer in enumerate(layers):
                stride = 1 if i == 0 else 2
                self.features.add(self._make_layer(block, num_layer, channels[i + 1],
                                                   stride, i + 1, in_channels=in_channels))
                in_channels = channels[i + 1]
            self.features.add(nn.BatchNorm())
            self.features.add(nn.PReLU())
            self.features.add(nn.GlobalAvgPool2D())
            self.features.add(nn.Flatten())

            self.features.add(nn.Dense(embedding_size, use_bias=False))
            self.features.add(nn.BatchNorm(scale=False, center=False))
            self.features.add(nn.PReLU())

    def _make_layer(self, block, layers, channels, stride, stage_index, in_channels=0):
        layer = nn.HybridSequential(prefix='stage%d_' % stage_index)
        with layer.name_scope():
            layer.add(block(channels, stride, channels != in_channels, in_channels=in_channels,
                            prefix=''))
            for _ in range(layers - 1):
                layer.add(block(channels, 1, False, in_channels=channels, prefix=''))
        return layer


resnet_spec = {18: (SE_BottleneckV2, [2, 2, 2, 2], [64, 64, 128, 256, 512]),
               34: (SE_BottleneckV2, [3, 4, 6, 3], [64, 64, 128, 256, 512]),
               50: (SE_BottleneckV2, [3, 4, 6, 3], [64, 256, 512, 1024, 2048]),
               101: (SE_BottleneckV2, [3, 4, 23, 3], [64, 256, 512, 1024, 2048]),
               152: (SE_BottleneckV2, [3, 8, 36, 3], [64, 256, 512, 1024, 2048])}


def get_se_resnet(num_layers, **kwargs):
    r"""SE_ResNet V1 model from `"Deep Residual Learning for Image Recognition"
    <http://arxiv.org/abs/1512.03385>`_ paper.
    SE_ResNet V2 model from `"Identity Mappings in Deep Residual Networks"
    <https://arxiv.org/abs/1603.05027>`_ paper.

    Parameters
    ----------
    version : int
        Version of ResNet. Options are 1, 2.
    num_layers : int
        Numbers of layers. Options are 18, 34, 50, 101, 152.
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.
    """
    assert num_layers in resnet_spec, \
        "Invalid number of layers: %d. Options are %s" % (
            num_layers, str(resnet_spec.keys()))

    block_type, layers, channels = resnet_spec[num_layers]
    net = SE_ResNetV2(block_type, layers, channels, **kwargs)
    return net


def se_resnet18_v2(**kwargs):
    r"""SE_ResNet-18 V2 model from `"Identity Mappings in Deep Residual Networks"
    <https://arxiv.org/abs/1603.05027>`_ paper.

    Parameters
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.
    """
    return get_se_resnet(18, **kwargs)


def se_resnet34_v2(**kwargs):
    r"""SE_ResNet-34 V2 model from `"Identity Mappings in Deep Residual Networks"
    <https://arxiv.org/abs/1603.05027>`_ paper.

    Parameters
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.
    """
    return get_se_resnet(34, **kwargs)


def se_resnet50_v2(**kwargs):
    r"""SE_ResNet-50 V2 model from `"Identity Mappings in Deep Residual Networks"
    <https://arxiv.org/abs/1603.05027>`_ paper.

    Parameters
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.
    """
    return get_se_resnet(50, **kwargs)


def se_resnet101_v2(**kwargs):
    r"""SE_ResNet-101 V2 model from `"Identity Mappings in Deep Residual Networks"
    <https://arxiv.org/abs/1603.05027>`_ paper.

    Parameters
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.
    """
    return get_se_resnet(101, **kwargs)


def se_resnet152_v2(**kwargs):
    r"""SE_ResNet-152 V2 model from `"Identity Mappings in Deep Residual Networks"
    <https://arxiv.org/abs/1603.05027>`_ paper.

    Parameters
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.
    """
    return get_se_resnet(152, **kwargs)




# MIT License
#
# Copyright (c) 2018 Haoxintong
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Basic Blocks used in GluonFR."""

__all__ = ['NormDense', 'SELayer', 'FrBase']

from mxnet.gluon import nn
from mxnet.gluon.nn import HybridBlock


class NormDense(HybridBlock):
    """Norm Dense"""
    def __init__(self, classes, weight_norm=False, feature_norm=False,
                 dtype='float32', weight_initializer=None, in_units=0, **kwargs):
        super().__init__(**kwargs)
        self._weight_norm = weight_norm
        self._feature_norm = feature_norm

        self._classes = classes
        self._in_units = in_units
        if weight_norm:
            assert in_units > 0, "Weight shape cannot be inferred auto when use weight norm, " \
                                 "in_units should be given."
        with self.name_scope():
            self.weight = self.params.get('weight', shape=(classes, in_units),
                                          init=weight_initializer, dtype=dtype,
                                          allow_deferred_init=True)

    # noinspection PyMethodOverriding
    def hybrid_forward(self, F, x, weight, *args, **kwargs):
        if self._weight_norm:
            weight = F.L2Normalization(weight, mode='instance')
        if self._feature_norm:
            x = F.L2Normalization(x, mode='instance', name='fc1n')
        return F.FullyConnected(data=x, weight=weight, no_bias=True,
                                num_hidden=self._classes, name='fc7')

    def __repr__(self):
        s = '{name}({layout})'
        shape = self.weight.shape
        return s.format(name=self.__class__.__name__,
                        layout='{0} -> {1}'.format(shape[1] if shape[1] else None, shape[0]))


class SELayer(HybridBlock):
    """SE Layer"""
    def __init__(self, channel, in_channel, reduction=16, **kwargs):
        super(SELayer, self).__init__(**kwargs)
        with self.name_scope():
            self.avg_pool = nn.GlobalAvgPool2D()
            self.fc = nn.HybridSequential()
            with self.fc.name_scope():
                self.fc.add(nn.Conv2D(channel // reduction, kernel_size=1, in_channels=in_channel))
                self.fc.add(nn.PReLU())
                self.fc.add(nn.Conv2D(channel, kernel_size=1, in_channels=channel // reduction))
                self.fc.add(nn.Activation('sigmoid'))

    def hybrid_forward(self, F, x, *args, **kwargs):
        y = self.avg_pool(x)
        y = self.fc(y)
        return F.broadcast_mul(x, y)


class FrBase(nn.HybridBlock):
    r"""
    This is base class for all face recognition network.
    In this class, we defined the NormDense and control flow of the sub classes.
    In any sub classes, only need to implement features and embedding_layer.
    Normally we add embedding_layer to features.

    Parameters
    ----------
    classes : int
        Number of classification classes.
    embedding_size : int
        Units of embedding layer.
    weight_norm : bool, default False
        Whether use weight norm in NormDense layer.
    feature_norm : bool, default False
        Whether use features norm in NormDense layer.
    need_cls_layer : bool, default True
        Whether use NormDense layer.Normally it depends on your loss function.
        When you use Softmax, ArcLoss or based on Softmax loss, you need to set it to True.
        When you only need embedding output, like you are predicting or training with triplet loss,
        you need to set it to False.
    """
    def __init__(self, classes=-1, embedding_size=512, weight_norm=False, feature_norm=False,
                 need_cls_layer=True, **kwargs):
        super(FrBase, self).__init__(**kwargs)
        self.need_cls_layer = need_cls_layer
        self._fn = feature_norm
        self.features = None

        if need_cls_layer:
            assert classes > 0, "When need_cls_layer=True, you need to specify a positive number of classes"
            self.output = NormDense(classes, weight_norm, feature_norm,
                                    in_units=embedding_size, prefix='output_')

    def hybrid_forward(self, F, x, *args, **kwargs):
        if self.features is None:
            raise NotImplementedError

        embedding = self.features(x)
        if self.need_cls_layer:
            out = self.output(embedding)
            return embedding, out
        else:
            if self._fn:
                embedding = F.L2Normalization(embedding, mode='instance')
                return embedding
            return embedding
